@@ -18,9 +18,10 @@ module Hledger.Data.Balancing
 , isTransactionBalanced
 , balanceTransaction
 , balanceTransactionHelper
+  -- * assertion validation
+, transactionCheckAssertions
   -- * journal balancing
 , journalBalanceTransactions
-, journalCheckBalanceAssertions
   -- * tests
 , tests_Balancing
 )
@@ -33,22 +34,22 @@ import Control.Monad.Reader as R (ReaderT, reader, runReaderT, ask, asks)
 import Control.Monad.ST (ST, runST)
 import Control.Monad.Trans.Class (lift)
 import Data.Array.ST (STArray, getElems, newListArray, writeArray)
+import Data.Bifunctor (second)
 import Data.Foldable (asum)
 import Data.Function ((&))
-import Data.Functor ((<&>))
-import "base-compat" Data.Functor.Compat (void)
-import qualified Data.HashTable.Class as H (toList)
-import qualified Data.HashTable.ST.Cuckoo as H
+import Data.Functor ((<&>), void)
+import Data.HashTable.Class qualified as H (toList)
+import Data.HashTable.ST.Cuckoo qualified as H
 import Data.List (partition, sortOn)
 import Data.List.Extra (nubSort)
 import Data.Maybe (fromJust, fromMaybe, isJust, isNothing, mapMaybe)
-import qualified Data.Set as S
-import qualified Data.Text as T
+import Data.Set qualified as S
+import Data.Text qualified as T
 import Data.Time.Calendar (fromGregorian)
-import qualified Data.Map as M
+import Data.Map qualified as M
+import Safe (headErr)
 import Text.Printf (printf)
 
-import Hledger.Utils
 import Hledger.Data.Types
 import Hledger.Data.AccountName (isAccountNamePrefixOf)
 import Hledger.Data.Amount
@@ -56,7 +57,7 @@ import Hledger.Data.Journal
 import Hledger.Data.Posting
 import Hledger.Data.Transaction
 import Hledger.Data.Errors
-import Data.Bifunctor (second)
+import Hledger.Utils
 
 
 data BalancingOpts = BalancingOpts
@@ -64,13 +65,15 @@ data BalancingOpts = BalancingOpts
   , infer_balancing_costs_ :: Bool  -- ^ Are we permitted to infer missing costs to balance transactions ?
                                     --   Distinct from InputOpts{infer_costs_}.
   , commodity_styles_      :: Maybe (M.Map CommoditySymbol AmountStyle)  -- ^ commodity display styles
-  } deriving (Show)
+  , txn_balancing_         :: TransactionBalancingPrecision
+  } deriving (Eq, Ord, Show)
 
 defbalancingopts :: BalancingOpts
 defbalancingopts = BalancingOpts
   { ignore_assertions_     = False
   , infer_balancing_costs_ = True
   , commodity_styles_      = Nothing
+  , txn_balancing_         = TBPExact
   }
 
 -- | Check that this transaction would appear balanced to a human when displayed.
@@ -91,7 +94,7 @@ defbalancingopts = BalancingOpts
 --    (using the given display styles if provided)
 --
 transactionCheckBalanced :: BalancingOpts -> Transaction -> [String]
-transactionCheckBalanced BalancingOpts{commodity_styles_} t = errs
+transactionCheckBalanced BalancingOpts{commodity_styles_=_mglobalstyles, txn_balancing_} t = errs
   where
     -- get real and balanced virtual postings, to be checked separately
     (rps, bvps) = foldr partitionPosting ([], []) $ tpostings t
@@ -101,45 +104,75 @@ transactionCheckBalanced BalancingOpts{commodity_styles_} t = errs
             BalancedVirtualPosting -> (l, p:r)
             VirtualPosting         -> (l, r)
 
-    -- convert this posting's amount to cost,
-    -- without getting confused by redundant costs/equity postings
+    -- convert a posting's amount to cost,
+    -- unless it has been marked as a redundant cost (equivalent to some nearby equity conversion postings),
+    -- in which case ignore it.
     postingBalancingAmount p
-      | "_price-matched" `elem` map fst (ptags p) = mixedAmountStripPrices $ pamount p
-      | otherwise                                 = mixedAmountCost $ pamount p
+      | costPostingTagName `elem` map fst (ptags p) = mixedAmountStripCosts $ pamount p
+      | otherwise                                   = mixedAmountCost $ pamount p
 
-    -- transaction balancedness is checked at each commodity's display precision
-    lookszero = mixedAmountLooksZero . atdisplayprecision
-      where
-        atdisplayprecision = maybe id styleAmounts $ commodity_styles_
+    lookszero = case txn_balancing_ of
+      TBPOld    -> lookszeroatglobaldisplayprecision
+      TBPExact  -> lookszeroatlocaltransactionprecision
+
+    lookszeroatlocaltransactionprecision = mixedAmountLooksZero . styleAmounts (transactionCommodityStylesWith HardRounding t)
+    lookszeroatglobaldisplayprecision    = mixedAmountLooksZero . maybe id styleAmounts _mglobalstyles
+
+    -- check that the sum looks like zero
+    (rsumcost,  bvsumcost)  = (foldMap postingBalancingAmount rps, foldMap postingBalancingAmount bvps)
+    (rsumok,    bvsumok)    = (lookszero rsumcost, lookszero bvsumcost)
+    (rsumokold, bvsumokold) = (lookszeroatglobaldisplayprecision rsumcost, lookszeroatglobaldisplayprecision bvsumcost)
 
     -- when there's multiple non-zeros, check they do not all have the same sign
-    (rsignsok, bvsignsok) = (signsOk rps, signsOk bvps)
+    (rsignsok, bvsignsok)   = (signsOk rps, signsOk bvps)
       where
         signsOk ps = length nonzeros < 2 || length nonzerosigns > 1
           where
             nonzeros = filter (not.lookszero) $ map postingBalancingAmount ps
             nonzerosigns = nubSort $ mapMaybe isNegativeMixedAmount nonzeros
 
-    -- check that the sum looks like zero
-    (rsumcost, bvsumcost) = (foldMap postingBalancingAmount rps, foldMap postingBalancingAmount bvps)
-    (rsumok, bvsumok) = (lookszero rsumcost, lookszero bvsumcost)
-
     -- Generate error messages if any. Show amounts with their original precisions.
     errs = filter (not.null) [rmsg, bvmsg]
       where
         rmsg
           | rsumok        = ""
           | not rsignsok  = "The real postings all have the same sign. Consider negating some of them."
-          | otherwise     = "The real postings' sum should be 0 but is: " ++ showMixedAmountOneLineWithoutPrice False rsumcost
+          | otherwise     = "The real postings' sum should be 0 but is: " ++
+              (showMixedAmountWith oneLineNoCostFmt{displayCost=True, displayZeroCommodity=True} $
+              mixedAmountSetFullPrecisionUpTo Nothing $ mixedAmountSetFullPrecision
+              rsumcost)
+              ++ if rsumokold then oldbalancingmsg else ""
         bvmsg
           | bvsumok       = ""
           | not bvsignsok = "The balanced virtual postings all have the same sign. Consider negating some of them."
-          | otherwise     = "The balanced virtual postings' sum should be 0 but is: " ++ showMixedAmountOneLineWithoutPrice False bvsumcost
+          | otherwise     = "The balanced virtual postings' sum should be 0 but is: " ++
+              (showMixedAmountWith oneLineNoCostFmt{displayCost=True, displayZeroCommodity=True} $
+              mixedAmountSetFullPrecisionUpTo Nothing $ mixedAmountSetFullPrecision
+              bvsumcost)
+              ++ if bvsumokold then oldbalancingmsg else ""
+        oldbalancingmsg = unlines [
+          -- -------------------------------------------------------------------------------
+           "\nNote, hledger <1.50 accepted this entry because of the global display precision,"
+          ,"but hledger 1.50+ checks more strictly, using the entry's local precision."
+          ,"You can use --txn-balancing=old to keep it working, or fix it (recommended);"
+          ,"see 'Transaction balancing' in the hledger manual."
+          ]
 
 -- | Legacy form of transactionCheckBalanced.
 isTransactionBalanced :: BalancingOpts -> Transaction -> Bool
 isTransactionBalanced bopts = null . transactionCheckBalanced bopts
 
+-- | Verify that any assertions in this transaction hold 
+-- when included in the larger journal.
+transactionCheckAssertions :: BalancingOpts -> Journal -> Transaction -> Either String Transaction
+transactionCheckAssertions bopts j t =
+  if (ignore_assertions_ bopts) then Right t else do
+    j' <- journalStyleAmounts j 
+    let newtxns = sortOn tdate (jtxns j' ++ [ t ])
+    case journalBalanceTransactions bopts j'{jtxns = newtxns} of
+      Right _ -> Right t
+      Left e -> Left e
+
 -- | Balance this transaction, ensuring that its postings
 -- (and its balanced virtual postings) sum to 0,
 -- by inferring a missing amount or conversion price(s) if needed.
@@ -188,9 +221,7 @@ balanceTransactionHelper bopts t = do
           [ "Consider adjusting this entry's amounts, adding missing postings,"
           , "or recording conversion price(s) with @, @@ or equity postings." 
           ]
-          else
-          [ "Consider adjusting this entry's amounts, or adding missing postings."
-          ]
+          else []
 
 transactionCommodities :: Transaction -> S.Set CommoditySymbol
 transactionCommodities t = mconcat $ map (maCommodities . pamount) $ tpostings t
@@ -270,7 +301,7 @@ transactionInferBalancingAmount styles t@Transaction{tpostings=ps}
                 & mixedAmountCost
                 -- & dbg9With (lbl "balancing amount converted to cost".showMixedAmountOneLine)
                 & styleAmounts (styles
-                                -- Needed until we switch to locally-inferred balancing precisions:
+                                -- Needed until we switch to locally-inferred balancing precisions: XXX #2402
                                 -- these had hard rounding set to help with balanced-checking;
                                 -- set no rounding now to avoid excessive display precision in output
                                 & amountStylesSetRounding NoRounding
@@ -339,7 +370,7 @@ costInferrerFor t pt = maybe id infercost inferFromAndTo
     inferFromAndTo = case sumamounts of
       [a,b] | noprices, oppositesigns -> asum $ map orderIfMatches pcommodities
         where
-          noprices      = all (isNothing . aprice) sumamounts
+          noprices      = all (isNothing . acost) sumamounts
           oppositesigns = signum (aquantity a) /= signum (aquantity b)
           orderIfMatches x | x == acommodity a = Just (a,b)
                            | x == acommodity b = Just (b,a)
@@ -351,29 +382,24 @@ costInferrerFor t pt = maybe id infercost inferFromAndTo
     -- then set its cost based on the ratio between fromamount and toamount.
     infercost (fromamount, toamount) p
         | [a] <- amounts (pamount p), ptype p == pt, acommodity a == acommodity fromamount
-            = p{ pamount   = mixedAmount a{aprice=Just conversionprice}
+            = p{ pamount   = mixedAmount a{acost=Just conversionprice}
                   & dbg9With (lbl "inferred cost".showMixedAmountOneLine)
                , poriginal = Just $ originalPosting p }
         | otherwise = p
       where
-        -- If only one Amount in the posting list matches fromamount we can use TotalPrice.
+        -- If only one Amount in the posting list matches fromamount we can use TotalCost.
         -- Otherwise divide the conversion equally among the Amounts by using a unit price.
         conversionprice = case filter (== acommodity fromamount) pcommodities of
-            [_] -> TotalPrice $ negate toamount
-            _   -> UnitPrice  $ negate unitprice `withPrecision` unitprecision
+            [_] -> TotalCost $ negate toamount
+            _   -> UnitCost  $ negate unitcost `withPrecision` unitprecision
 
-        unitprice     = aquantity fromamount `divideAmount` toamount
+        unitcost     = aquantity fromamount `divideAmount` toamount
         unitprecision = case (asprecision $ astyle fromamount, asprecision $ astyle toamount) of
             (Precision a, Precision b) -> Precision . max 2 $ saturatedAdd a b
             _                          -> NaturalPrecision
         saturatedAdd a b = if maxBound - a < b then maxBound else a + b
 
 
--- | Check any balance assertions in the journal and return an error message
--- if any of them fail (or if the transaction balancing they require fails).
-journalCheckBalanceAssertions :: Journal -> Maybe String
-journalCheckBalanceAssertions = either Just (const Nothing) . journalBalanceTransactions defbalancingopts
-
 -- "Transaction balancing", including: inferring missing amounts,
 -- applying balance assignments, checking transaction balancedness,
 -- checking balance assertions, respecting posting dates. These things
@@ -526,7 +552,7 @@ journalBalanceTransactions bopts' j' =
         ts' <- lift $ getElems balancedtxns
         return j{jtxns=ts'}
 
--- Before #2039: "Costs are removed, which helps eg balance-assertions.test: 15. Mix different commodities and assignments."
+-- Before #2039: "Costs are removed, which helps eg assertions.test: 15. Mix different commodities and assignments."
 
 -- | This function is called statefully on each of a date-ordered sequence of
 -- 1. fully explicit postings from already-balanced transactions and
@@ -539,8 +565,8 @@ journalBalanceTransactions bopts' j' =
 balanceTransactionAndCheckAssertionsB :: Either Posting Transaction -> Balancing s ()
 balanceTransactionAndCheckAssertionsB (Left p@Posting{}) =
   -- Update the account's running balance and check the balance assertion if any.
-  -- Note, cost is ignored when checking balance assertions, currently.
-  void . addAmountAndCheckAssertionB $ postingStripPrices p
+  -- Cost is ignored when checking balance assertions currently.
+  void $ addAmountAndCheckAssertionB $ postingStripCosts p
 balanceTransactionAndCheckAssertionsB (Right t@Transaction{tpostings=ps}) = do
   -- make sure we can handle the balance assignments
   mapM_ checkIllegalBalanceAssignmentB ps
@@ -670,7 +696,7 @@ checkBalanceAssertionOneCommodityB p@Posting{paccount=assertedacct} assertedcomm
         "%s\n",
         "Balance assertion failed in %s",
         "%s at this point, %s, ignoring costs,",
-        "the expected balance is:        %s",
+        "the asserted balance is:        %s",
         "but the calculated balance is:  %s",
         "(difference: %s)",
         "To troubleshoot, check this account's running balance with assertions disabled, eg:",
@@ -682,9 +708,12 @@ checkBalanceAssertionOneCommodityB p@Posting{paccount=assertedacct} assertedcomm
       acct  -- asserted account
       (if istotal then "Across all commodities" else "In commodity " <> assertedcommstr)  -- asserted commodity or all commodities ?
       (if isinclusive then "including subaccounts" else "excluding subaccounts" :: String)  -- inclusive or exclusive balance asserted ?
-      (pad assertedstr)  -- asserted amount, without cost
-      (pad actualstr)    -- actual amount, without cost
-        -- <> " (with costs: " <> T.pack (showMixedAmountWith fmt actualcommbal) <> ")" -- debugging
+      (pad assertedstr  -- asserted amount, without cost
+        <> if debugLevel >= 2 then " (with cost:  " <> T.pack (showAmountWith fmt assertedcommbal) <> ")" else ""
+      )
+      (pad actualstr    -- actual amount, without cost
+        <> if debugLevel >= 2 then " (with costs: " <> T.pack (showMixedAmountWith fmt actualcommbal) <> ")" else ""
+      )
       diffstr  -- their difference
       (acct ++ if isinclusive then "" else "$")  -- query matching the account(s) postings
       (if istotal then "" else (" cur:" ++ quoteForCommandLine (T.unpack assertedcomm)))  -- query matching the commodity(ies)
@@ -695,7 +724,7 @@ checkBalanceAssertionOneCommodityB p@Posting{paccount=assertedacct} assertedcomm
         pos = baposition ass
         (_,_,_,ex) = makeBalanceAssertionErrorExcerpt p
         assertedcommstr = if T.null assertedcomm then "\"\"" else assertedcomm
-        fmt = oneLine{displayZeroCommodity=True}
+        fmt = oneLineFmt{displayZeroCommodity=True}
         assertedstr = showAmountWith fmt assertedcommbalcostless
         actualstr   = showAmountWith fmt actualcommbalcostless
         diffstr     = showAmountWith fmt $ assertedcommbalcostless - actualcommbalcostless
@@ -788,7 +817,7 @@ tests_Balancing =
                (Transaction
                   0
                   ""
-                  nullsourcepos
+                  nullsourcepospair
                   (fromGregorian 2007 01 28)
                   Nothing
                   Unmarked
@@ -803,7 +832,7 @@ tests_Balancing =
                (Transaction
                   0
                   ""
-                  nullsourcepos
+                  nullsourcepospair
                   (fromGregorian 2007 01 28)
                   Nothing
                   Unmarked
@@ -820,7 +849,7 @@ tests_Balancing =
              (Transaction
                 0
                 ""
-                nullsourcepos
+                nullsourcepospair
                 (fromGregorian 2007 01 28)
                 Nothing
                 Unmarked
@@ -831,12 +860,12 @@ tests_Balancing =
                 [posting {paccount = "a", pamount = mixedAmount (usd 1)}, posting {paccount = "b", pamount = missingmixedamt}])) @?=
           Right (mixedAmount $ usd (-1))
         ,testCase "conversion price is inferred" $
-          (pamount . head . tpostings <$>
+          (pamount . headErr . tpostings <$>  -- PARTIAL headErr succeeds because non-null postings list
            balanceTransaction defbalancingopts
              (Transaction
                 0
                 ""
-                nullsourcepos
+                nullsourcepospair
                 (fromGregorian 2007 01 28)
                 Nothing
                 Unmarked
@@ -854,7 +883,7 @@ tests_Balancing =
             (Transaction
                0
                ""
-               nullsourcepos
+               nullsourcepospair
                (fromGregorian 2011 01 01)
                Nothing
                Unmarked
@@ -871,7 +900,7 @@ tests_Balancing =
             (Transaction
                0
                ""
-               nullsourcepos
+               nullsourcepospair
                (fromGregorian 2011 01 01)
                Nothing
                Unmarked
@@ -890,7 +919,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -908,7 +937,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -926,7 +955,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -941,7 +970,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -956,7 +985,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -975,7 +1004,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -993,7 +1022,7 @@ tests_Balancing =
           Transaction
             0
             ""
-            nullsourcepos
+            nullsourcepospair
             (fromGregorian 2009 01 01)
             Nothing
             Unmarked
@@ -1023,7 +1052,7 @@ tests_Balancing =
               transaction (fromGregorian 2019 01 01) [ vpost' "a" missingamt (balassert (num 1)) ]
             ]}
       assertRight ej
-      case ej of Right j -> (jtxns j & head & tpostings & head & pamount & amountsRaw) @?= [num 1]
+      case ej of Right j -> (jtxns j & headErr & tpostings & headErr & pamount & amountsRaw) @?= [num 1]  -- PARTIAL headErrs succeed because non-null txns & postings lists given
                  Left _  -> error' "balance-assignment test: shouldn't happen"
 
     ,testCase "same-day-1" $ do
@@ -1068,6 +1097,32 @@ tests_Balancing =
 
     ]
 
+    ,testGroup "transactionCheckAssertions" $ [
+      testCase "simple assertion on same day" $ do 
+        assertRight $ 
+          transactionCheckAssertions defbalancingopts nulljournal{ jtxns = [
+            transaction (fromGregorian 2025 01 01) [ vpost' "a" (usd 1) Nothing ]
+            ] } (transaction (fromGregorian 2025 01 01) [ vpost' "a" (usd 1) (balassert (usd 2)) ])
+
+      ,testCase "inclusive assertions" $ do 
+        assertRight $ 
+          transactionCheckAssertions defbalancingopts nulljournal{ jtxns = [
+            transaction (fromGregorian 2025 01 01) [ vpost' "a:a" (usd 1) Nothing ] 
+            ,transaction (fromGregorian 2025 01 02) [ vpost' "a:b" (usd 2) Nothing]
+            ,transaction (fromGregorian 2025 01 02) [ vpost' "a:c" (usd 5) Nothing]
+            ,transaction (fromGregorian 2025 01 03) [ vpost' "a:d" (eur 10) Nothing]
+            ] } (transaction (fromGregorian 2025 01 04) [ vpost' "a" (usd 2) (balassertParInc (usd 10))])
+
+      ,testCase "multicommodity assertion" $ do 
+        assertRight $ 
+          transactionCheckAssertions defbalancingopts nulljournal{ jtxns = [
+            transaction (fromGregorian 2025 01 01) [ vpost' "a" (usd 1) Nothing ] 
+            ,transaction (fromGregorian 2025 01 02) [ vpost' "a:b" (eur 2) Nothing ]
+            ,transaction (fromGregorian 2025 01 02) [ vpost' "a:c" (usd 5) Nothing ]
+            ,transaction (fromGregorian 2025 01 03) [ vpost' "a:b" (eur (-2)) Nothing ]
+            ] } (transaction (fromGregorian 2025 01 03) [ vpost' "a" (usd 2) (balassertTotInc (usd 8)) ])
+    ]
+
     ,testGroup "commodityStylesFromAmounts" $ [
 
       -- Journal similar to the one on #1091:
