@@ -1,9 +1,10 @@
+{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE Rank2Types          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections       #-}
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE RecordWildCards     #-}
 
 {-|
 
@@ -21,16 +22,15 @@ module Hledger.Data.Journal (
   addTransactionModifier,
   addPeriodicTransaction,
   addTransaction,
+  journalDbg,
   journalInferMarketPricesFromTransactions,
   journalInferCommodityStyles,
   journalStyleAmounts,
-  commodityStylesFromAmounts,
   journalCommodityStyles,
   journalCommodityStylesWith,
   journalToCost,
   journalInferEquityFromCosts,
-  journalInferCostsFromEquity,
-  journalMarkRedundantCosts,
+  journalTagCostsAndEquityAndMaybeInferCosts,
   journalReverse,
   journalSetLastReadTime,
   journalRenumberAccountDeclarations,
@@ -55,7 +55,6 @@ module Hledger.Data.Journal (
   journalAccountNamesDeclared,
   journalAccountNamesDeclaredOrUsed,
   journalAccountNamesDeclaredOrImplied,
-  journalLeafAccountNamesDeclared,
   journalAccountNames,
   journalLeafAccountNames,
   journalAccountNameTree,
@@ -72,8 +71,14 @@ module Hledger.Data.Journal (
   journalTagsDeclared,
   journalTagsUsed,
   journalTagsDeclaredOrUsed,
+  journalAmounts,
+  journalPostingAmounts,
+  journalPostingAndCostAmounts,
   journalCommoditiesDeclared,
+  journalCommoditiesUsed,
   journalCommodities,
+  journalCommoditiesFromPriceDirectives,
+  journalCommoditiesFromTransactions,
   journalDateSpan,
   journalDateSpanBothDates,
   journalStartDate,
@@ -86,19 +91,19 @@ module Hledger.Data.Journal (
   journalNextTransaction,
   journalPrevTransaction,
   journalPostings,
-  journalPostingAmounts,
-  showJournalAmountsDebug,
+  showJournalPostingAmountsDebug,
   journalTransactionsSimilarTo,
   -- * Account types
   journalAccountType,
   journalAccountTypes,
   journalAddAccountTypes,
   journalPostingsAddAccountTags,
+  journalPostingsKeepAccountTagsOnly,
+  defaultBaseConversionAccount,
   -- journalPrices,
-  journalConversionAccount,
+  journalBaseConversionAccount,
   journalConversionAccounts,
   -- * Misc
-  canonicalStyleFrom,
   nulljournal,
   journalConcat,
   journalNumberTransactions,
@@ -110,7 +115,7 @@ module Hledger.Data.Journal (
   -- * Tests
   samplejournal,
   samplejournalMaybeExplicit,
-  tests_Journal
+  tests_Journal,
   --
 )
 where
@@ -121,20 +126,22 @@ import Control.Monad.State.Strict (StateT)
 import Data.Char (toUpper, isDigit)
 import Data.Default (Default(..))
 import Data.Foldable (toList)
-import Data.List ((\\), find, foldl', sortBy, union, intercalate)
+import Data.List ((\\), find, sortBy, union, intercalate)
+#if !MIN_VERSION_base(4,20,0)
+import Data.List (foldl')
+#endif
 import Data.List.Extra (nubSort)
-import qualified Data.Map.Strict as M
+import Data.Map.Strict qualified as M
 import Data.Maybe (catMaybes, fromMaybe, mapMaybe, maybeToList)
-import qualified Data.Set as S
+import Data.Set qualified as S
 import Data.Text (Text)
-import qualified Data.Text as T
+import Data.Text qualified as T
 import Safe (headMay, headDef, maximumMay, minimumMay, lastDef)
 import Data.Time.Calendar (Day, addDays, fromGregorian, diffDays)
 import Data.Time.Clock.POSIX (POSIXTime)
 import Data.Tree (Tree(..), flatten)
 import Text.Printf (printf)
 import Text.Megaparsec (ParsecT)
-import Text.Megaparsec.Custom (FinalParseError)
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -149,6 +156,7 @@ import System.FilePath (takeFileName)
 import Data.Ord (comparing)
 import Hledger.Data.Dates (nulldate)
 import Data.List (sort)
+import Data.Function ((&))
 -- import Data.Function ((&))
 
 
@@ -178,21 +186,43 @@ instance Show Journal where
              (length $ jtxns j)
              (length accounts)
              (show accounts)
-             (show $ jinferredcommodities j)
+             (show $ jinferredcommoditystyles j)
              -- ++ (show $ journalTransactions l)
              where accounts = filter (/= "root") $ flatten $ journalAccountNameTree j
 
--- showJournalDebug j = unlines [
---                       show j
---                      ,show (jtxns j)
---                      ,show (jtxnmodifiers j)
---                      ,show (jperiodictxns j)
---                      ,show $ jparsetimeclockentries j
---                      ,show $ jpricedirectives j
---                      ,show $ jfinalcommentlines j
---                      ,show $ jparsestate j
---                      ,show $ map fst $ jfiles j
---                      ]
+journalDbg j@Journal{..} = chomp $ unlines $
+  ("Journal " ++ takeFileName (journalFilePath j)++":") :  --  ++ " {"
+  map (" "<>) [
+   "jparsedefaultyear: "         <> shw jparsedefaultyear
+  ,"jparsedefaultcommodity: "    <> shw jparsedefaultcommodity
+  ,"jparsedecimalmark: "         <> shw jparsedecimalmark
+  ,"jparseparentaccounts: "      <> shw jparseparentaccounts
+  ,"jparsealiases: "             <> shw jparsealiases
+  -- ,"jparsetimeclockentries: " <> shw jparsetimeclockentries
+  ,"jincludefilestack: "         <> shw jincludefilestack
+  ,"jdeclaredpayees: "           <> shw jdeclaredpayees
+  ,"jdeclaredtags: "             <> shw jdeclaredtags
+  ,"jdeclaredaccounts: "         <> shw jdeclaredaccounts
+  ,"jdeclaredaccounttags: "      <> shw jdeclaredaccounttags
+  ,"jdeclaredaccounttypes: "     <> shw jdeclaredaccounttypes
+  ,"jaccounttypes: "             <> shw jaccounttypes
+  ,"jdeclaredcommodities: "      <> shw jdeclaredcommodities
+  ,"jinferredcommoditystyles: "  <> shw jinferredcommoditystyles
+  ,"jglobalcommoditystyles: "    <> shw jglobalcommoditystyles
+  ,"jpricedirectives: "          <> shw jpricedirectives
+  ,"jinferredmarketprices: "     <> shw jinferredmarketprices
+  ,"jtxnmodifiers: "             <> shw jtxnmodifiers
+  -- ,"jperiodictxns: "          <> shw jperiodictxns
+  ,"jtxns: "                     <> shw jtxns
+  ,"jfinalcommentlines: "        <> shw jfinalcommentlines
+  ,"jfiles: "                    <> shw jfiles
+  ,"jlastreadtime: "             <> shw jlastreadtime
+  ]
+  -- ++ ["}"]
+  where
+    shw :: Show a => a -> String
+    shw = show
+    -- shw = pshow
 
 -- The semigroup instance for Journal is useful for two situations.
 --
@@ -232,12 +262,32 @@ journalConcat j1 j2 =
     ,jdeclaredpayees            = jdeclaredpayees            j1 <> jdeclaredpayees            j2
     ,jdeclaredtags              = jdeclaredtags              j1 <> jdeclaredtags              j2
     ,jdeclaredaccounts          = jdeclaredaccounts          j1 <> jdeclaredaccounts          j2
-    ,jdeclaredaccounttags       = jdeclaredaccounttags       j1 <> jdeclaredaccounttags       j2
-    ,jdeclaredaccounttypes      = jdeclaredaccounttypes      j1 <> jdeclaredaccounttypes      j2
-    ,jaccounttypes              = jaccounttypes              j1 <> jaccounttypes              j2
-    ,jglobalcommoditystyles     = jglobalcommoditystyles     j1 <> jglobalcommoditystyles     j2
-    ,jcommodities               = jcommodities               j1 <> jcommodities               j2
-    ,jinferredcommodities       = jinferredcommodities       j1 <> jinferredcommodities       j2
+    --
+    -- The next six fields are Maps, which need to be merged carefully for correct semantics,
+    -- especially the first two, which have list values. There may still be room for improvement here.
+    --
+    -- ,jdeclaredaccounttags   :: M.Map AccountName [Tag]
+    -- jdeclaredaccounttags can have multiple duplicated/conflicting values for an account's tag.
+    ,jdeclaredaccounttags       = M.unionWith (<>) (jdeclaredaccounttags j1) (jdeclaredaccounttags j2)
+    --
+    -- ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]
+    -- jdeclaredaccounttypes can have multiple duplicated/conflicting values for an account's type.
+    ,jdeclaredaccounttypes      = M.unionWith (<>) (jdeclaredaccounttypes j1) (jdeclaredaccounttypes j2)
+    --
+    -- ,jaccounttypes          :: M.Map AccountName AccountType
+    -- jaccounttypes has a single type for any given account. When it had multiple type declarations, the last/rightmost wins.
+    ,jaccounttypes              = M.unionWith (const id) (jaccounttypes j1) (jaccounttypes j2)
+    --
+    -- ,jglobalcommoditystyles :: M.Map CommoditySymbol AmountStyle
+    ,jglobalcommoditystyles     = (<>) (jglobalcommoditystyles j1) (jglobalcommoditystyles j2)
+    --
+    -- ,jdeclaredcommodities           :: M.Map CommoditySymbol Commodity
+    ,jdeclaredcommodities               = (<>) (jdeclaredcommodities j1) (jdeclaredcommodities j2)
+    --
+    -- ,jinferredcommoditystyles   :: M.Map CommoditySymbol AmountStyle
+    ,jinferredcommoditystyles       = (<>) (jinferredcommoditystyles j1) (jinferredcommoditystyles j2)
+    --
+    --
     ,jpricedirectives           = jpricedirectives           j1 <> jpricedirectives           j2
     ,jinferredmarketprices      = jinferredmarketprices      j1 <> jinferredmarketprices      j2
     ,jtxnmodifiers              = jtxnmodifiers              j1 <> jtxnmodifiers              j2
@@ -259,10 +309,10 @@ journalRenumberAccountDeclarations j = j{jdeclaredaccounts=jdas'}
     -- it seems unneeded except perhaps for debugging
 
 -- | Debug log the ordering of a journal's account declarations
--- (at debug level 5+).
+-- (at debug level 7+).
 dbgJournalAcctDeclOrder :: String -> Journal -> Journal
 dbgJournalAcctDeclOrder prefix =
-  traceOrLogAtWith 5 ((prefix++) . showAcctDeclsSummary . jdeclaredaccounts)
+  dbg7With ((prefix++) . showAcctDeclsSummary . jdeclaredaccounts)
   where
     showAcctDeclsSummary :: [(AccountName,AccountDeclarationInfo)] -> String
     showAcctDeclsSummary adis
@@ -295,8 +345,8 @@ nulljournal = Journal {
   ,jdeclaredaccounttypes      = M.empty
   ,jaccounttypes              = M.empty
   ,jglobalcommoditystyles     = M.empty
-  ,jcommodities               = M.empty
-  ,jinferredcommodities       = M.empty
+  ,jdeclaredcommodities               = M.empty
+  ,jinferredcommoditystyles       = M.empty
   ,jpricedirectives           = []
   ,jinferredmarketprices      = []
   ,jtxnmodifiers              = []
@@ -350,17 +400,47 @@ journalPostings = concatMap tpostings . jtxns
 journalPostingAmounts :: Journal -> [MixedAmount]
 journalPostingAmounts = map pamount . journalPostings
 
--- | Show the journal amounts rendered, suitable for debug logging.
-showJournalAmountsDebug :: Journal -> String
-showJournalAmountsDebug = show.map showMixedAmountOneLine.journalPostingAmounts
+-- | Show the journal posting amounts rendered, suitable for debug logging.
+showJournalPostingAmountsDebug :: Journal -> String
+showJournalPostingAmountsDebug = show . map showMixedAmountOneLine . journalPostingAmounts
+
+-- | All raw amounts used in this journal's postings and costs,
+-- with MixedAmounts flattened, in parse order.
+journalPostingAndCostAmounts :: Journal -> [Amount]
+journalPostingAndCostAmounts = concatMap getAmounts . concatMap (amountsRaw . pamount) . journalPostings
+
+-- | All raw amounts appearing in this journal, with MixedAmounts flattened, in no particular order.
+-- (Including from posting amounts, cost amounts, P directives, and the last D directive.)
+journalAmounts :: Journal -> S.Set Amount
+journalAmounts = S.fromList . journalStyleInfluencingAmounts True
 
 -- | Sorted unique commodity symbols declared by commodity directives in this journal.
 journalCommoditiesDeclared :: Journal -> [CommoditySymbol]
-journalCommoditiesDeclared = M.keys . jcommodities
+journalCommoditiesDeclared = M.keys . jdeclaredcommodities
 
--- | Sorted unique commodity symbols declared or inferred from this journal.
+-- | Sorted unique commodity symbols used anywhere in this journal, including
+-- commodity directives, P directives, the last D directive, posting amounts and cost amounts.
+journalCommoditiesUsed :: Journal -> [CommoditySymbol]
+journalCommoditiesUsed j = S.elems $
+  journalCommoditiesFromPriceDirectives j <>
+  (S.fromList $ map acommodity $ journalStyleInfluencingAmounts True j)
+
+-- | Sorted unique commodity symbols mentioned anywhere in this journal.
+-- (Including commodity directives, P directives, the last D directive, posting amounts and cost amounts.)
 journalCommodities :: Journal -> S.Set CommoditySymbol
-journalCommodities j = M.keysSet (jcommodities j) <> M.keysSet (jinferredcommodities j)
+journalCommodities j =
+     M.keysSet (jdeclaredcommodities j)
+  <> journalCommoditiesFromPriceDirectives j
+  <> S.fromList (map acommodity $ journalStyleInfluencingAmounts True j)
+
+-- | Sorted unique commodity symbols mentioned in this journal's P directives.
+journalCommoditiesFromPriceDirectives :: Journal -> S.Set CommoditySymbol
+journalCommoditiesFromPriceDirectives = S.fromList . concatMap pdcomms . jpricedirectives
+  where pdcomms pd = [pdcommodity pd, acommodity $ pdamount pd]
+
+-- | Sorted unique commodity symbols used in transactions, in either posting or cost amounts.
+journalCommoditiesFromTransactions :: Journal -> S.Set CommoditySymbol
+journalCommoditiesFromTransactions j = S.fromList $ map acommodity $ journalPostingAndCostAmounts j
 
 -- | Unique transaction descriptions used in this journal.
 journalDescriptions :: Journal -> [Text]
@@ -406,11 +486,6 @@ journalAccountNamesImplied = expandAccountNames . journalAccountNamesUsed
 journalAccountNamesDeclared :: Journal -> [AccountName]
 journalAccountNamesDeclared = nubSort . map fst . jdeclaredaccounts
 
--- | Sorted unique account names declared by account directives in this journal,
--- which have no children.
-journalLeafAccountNamesDeclared :: Journal -> [AccountName]
-journalLeafAccountNamesDeclared = treeLeaves . accountNameTreeFrom . journalAccountNamesDeclared
-
 -- | Sorted unique account names declared by account directives or posted to
 -- by transactions in this journal.
 journalAccountNamesDeclaredOrUsed :: Journal -> [AccountName]
@@ -529,29 +604,34 @@ journalAccountType Journal{jaccounttypes} = accountNameType jaccounttypes
 journalAddAccountTypes :: Journal -> Journal
 journalAddAccountTypes j = j{jaccounttypes = journalAccountTypes j}
 
+-- | An account type inherited from the parent account(s),
+-- and whether it was originally declared by an account directive (true) or inferred from an account name (false).
+type ParentAccountType = (AccountType, Bool)
+
 -- | Build a map of all known account types, explicitly declared
 -- or inferred from the account's parent or name.
 journalAccountTypes :: Journal -> M.Map AccountName AccountType
 journalAccountTypes j = M.fromList [(a,acctType) | (a, Just (acctType,_)) <- flatten t']
   where
     t = accountNameTreeFrom $ journalAccountNames j :: Tree AccountName
-    -- Map from the top of the account tree down to the leaves, propagating
-    -- account types downward. Keep track of whether the account is declared
-    -- (True), in which case the parent account should be preferred, or merely
-    -- inferred (False), in which case the inferred type should be preferred.
-    t' = settypes Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
+    -- Traverse downward through the account tree, applying any explicitly declared account types,
+    -- otherwise inferring account types from account names when possible, and propagating account types downward.
+    -- Declared account types (possibly inherited from parent) are preferred, inferred types are used as a fallback.
+    t' = setTypeHereAndBelow Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
       where
-        settypes :: Maybe (AccountType, Bool) -> Tree AccountName -> Tree (AccountName, Maybe (AccountType, Bool))
-        settypes mparenttype (Node a subs) = Node (a, mtype) (map (settypes mtype) subs)
+        declaredtypes       = M.keys $ jdeclaredaccounttypes j
+        declaredtypesbyname = journalDeclaredAccountTypes j & fmap (,True)
+        setTypeHereAndBelow :: Maybe ParentAccountType -> Tree AccountName -> Tree (AccountName, Maybe ParentAccountType)
+        setTypeHereAndBelow mparenttype (Node a subs) = Node (a, mnewtype) (map (setTypeHereAndBelow mnewtype) subs)
           where
-            mtype = M.lookup a declaredtypes <|> minferred
-              where 
-                declaredtypes = (,True) <$> journalDeclaredAccountTypes j
-                minferred = if maybe False snd mparenttype
-                            then mparenttype
-                            else (,False) <$> accountNameInferType a <|> mparenttype
+            mnewtype = mthisacctdeclaredtype <|> mparentacctdeclaredtype <|> mthisacctinferredtype <|> mparentacctinferredtype
+              where
+                mthisacctdeclaredtype   = M.lookup a declaredtypesbyname
+                mparentacctdeclaredtype = if       fromMaybe False $ snd <$> mparenttype then mparenttype else Nothing
+                mparentacctinferredtype = if not $ fromMaybe True  $ snd <$> mparenttype then mparenttype else Nothing
+                mthisacctinferredtype   = accountNameInferTypeExcept declaredtypes a & fmap (,False)  -- XXX not sure about this Except logic.. but for now, tests pass
 
--- | Build a map of the account types explicitly declared for each account.
+-- | Build a map from account names to explicitly declared account types.
 journalDeclaredAccountTypes :: Journal -> M.Map AccountName AccountType
 journalDeclaredAccountTypes Journal{jdeclaredaccounttypes} =
   M.fromList $ concat [map (,t) as | (t,as) <- M.toList jdeclaredaccounttypes]
@@ -563,18 +643,24 @@ journalPostingsAddAccountTags :: Journal -> Journal
 journalPostingsAddAccountTags j = journalMapPostings addtags j
   where addtags p = p `postingAddTags` (journalInheritedAccountTags j $ paccount p)
 
--- | The account to use for automatically generated conversion postings in this journal:
--- the first of the journalConversionAccounts.
-journalConversionAccount :: Journal -> AccountName
-journalConversionAccount = headDef defaultConversionAccount . journalConversionAccounts
+-- | Remove all tags from the journal's postings except those provided by their account.
+-- This is useful for the accounts report.
+-- It does not remove tag declarations from the posting comments.
+journalPostingsKeepAccountTagsOnly :: Journal -> Journal
+journalPostingsKeepAccountTagsOnly j = journalMapPostings keepaccounttags j
+  where keepaccounttags p = p{ptags=[]} `postingAddTags` (journalInheritedAccountTags j $ paccount p)
 
--- | All the accounts declared or inferred as Conversion type in this journal.
+-- | The account name to use for conversion postings generated by --infer-equity.
+-- This is the first account declared with type V/Conversion,
+-- or otherwise the defaultBaseConversionAccount (equity:conversion).
+journalBaseConversionAccount :: Journal -> AccountName
+journalBaseConversionAccount = headDef defaultBaseConversionAccount . journalConversionAccounts
+
+-- | All the accounts in this journal which are declared or inferred as V/Conversion type.
+-- This does not include new account names which might be generated by --infer-equity, currently.
 journalConversionAccounts :: Journal -> [AccountName]
 journalConversionAccounts = M.keys . M.filter (==Conversion) . jaccounttypes
 
--- The fallback account to use for automatically generated conversion postings
--- if no account is declared with the Conversion type.
-defaultConversionAccount = "equity:conversion"
 
 -- Various kinds of filtering on journals. We do it differently depending
 -- on the command.
@@ -839,9 +925,9 @@ journalCommodityStyles j =
   globalstyles <> declaredstyles <> defaultcommoditystyle <> inferredstyles
   where
     globalstyles          = jglobalcommoditystyles j
-    declaredstyles        = M.mapMaybe cformat $ jcommodities j
+    declaredstyles        = M.mapMaybe cformat $ jdeclaredcommodities j
     defaultcommoditystyle = M.fromList $ catMaybes [jparsedefaultcommodity j]
-    inferredstyles        = jinferredcommodities j
+    inferredstyles        = jinferredcommoditystyles j
 
 -- | Like journalCommodityStyles, but attach a particular rounding strategy to the styles,
 -- affecting how they will affect display precisions when applied.
@@ -849,59 +935,13 @@ journalCommodityStylesWith :: Rounding -> Journal -> M.Map CommoditySymbol Amoun
 journalCommodityStylesWith r = amountStylesSetRounding r . journalCommodityStyles
 
 -- | Collect and save inferred amount styles for each commodity based on
--- the posting amounts in that commodity (excluding price amounts), ie:
--- "the format of the first amount, adjusted to the highest precision of all amounts".
+-- P directive amounts, posting amounts but not cost amounts, and maybe the last D amount, in that commodity.
 -- Can return an error message eg if inconsistent number formats are found.
 journalInferCommodityStyles :: Journal -> Either String Journal
 journalInferCommodityStyles j =
-  case commodityStylesFromAmounts $ journalStyleInfluencingAmounts j of
+  case commodityStylesFromAmounts $ journalStyleInfluencingAmounts False j of
     Left e   -> Left e
-    Right cs -> Right j{jinferredcommodities = dbg7 "journalInferCommodityStyles" cs}
-
--- | Given a list of amounts, in parse order (roughly speaking; see journalStyleInfluencingAmounts),
--- build a map from their commodity names to standard commodity
--- display formats. Can return an error message eg if inconsistent
--- number formats are found.
---
--- Though, these amounts may have come from multiple files, so we
--- shouldn't assume they use consistent number formats.
--- Currently we don't enforce that even within a single file,
--- and this function never reports an error.
---
-commodityStylesFromAmounts :: [Amount] -> Either String (M.Map CommoditySymbol AmountStyle)
-commodityStylesFromAmounts =
-    Right . foldr (\a -> M.insertWith canonicalStyle (acommodity a) (astyle a)) mempty
-
--- | Given a list of amount styles (assumed to be from parsed amounts
--- in a single commodity), in parse order, choose a canonical style.
-canonicalStyleFrom :: [AmountStyle] -> AmountStyle
-canonicalStyleFrom = foldl' canonicalStyle amountstyle
-
--- TODO: should probably detect and report inconsistencies here.
--- Though, we don't have the info for a good error message, so maybe elsewhere.
--- | Given a pair of AmountStyles, choose a canonical style.
--- This is:
--- the general style of the first amount,
--- with the first digit group style seen,
--- with the maximum precision of all.
-canonicalStyle :: AmountStyle -> AmountStyle -> AmountStyle
-canonicalStyle a b = a{asprecision=prec, asdecimalmark=decmark, asdigitgroups=mgrps}
-  where
-    -- precision is maximum of all precisions
-    prec = max (asprecision a) (asprecision b)
-    -- identify the digit group mark (& group sizes)
-    mgrps = asdigitgroups a <|> asdigitgroups b
-    -- if a digit group mark was identified above, we can rely on that;
-    -- make sure the decimal mark is different. If not, default to period.
-    defdecmark = case mgrps of
-        Just (DigitGroups '.' _) -> ','
-        _                        -> '.'
-    -- identify the decimal mark: the first one used, or the above default,
-    -- but never the same character as the digit group mark.
-    -- urgh.. refactor..
-    decmark = case mgrps of
-        Just _  -> Just defdecmark
-        Nothing -> asdecimalmark a <|> asdecimalmark b <|> Just defdecmark
+    Right cs -> Right j{jinferredcommoditystyles = dbg7 "journalInferCommodityStyles" cs}
 
 -- -- | Apply this journal's historical price records to unpriced amounts where possible.
 -- journalApplyPriceDirectives :: Journal -> Journal
@@ -913,7 +953,7 @@ canonicalStyle a b = a{asprecision=prec, asdecimalmark=decmark, asdigitgroups=mg
 --         fixmixedamount = mapMixedAmount fixamount
 --         fixamount = fixprice
 --         fixprice a@Amount{price=Just _} = a
---         fixprice a@Amount{commodity=c} = a{price=maybe Nothing (Just . UnitPrice) $ journalPriceDirectiveFor j d c}
+--         fixprice a@Amount{commodity=c} = a{price=maybe Nothing (Just . UnitCost) $ journalPriceDirectiveFor j d c}
 
 -- -- | Get the price for a commodity on the specified day from the price database, if known.
 -- -- Does only one lookup step, ie will not look up the price of a price.
@@ -929,7 +969,7 @@ canonicalStyle a b = a{asprecision=prec, asdecimalmark=decmark, asdigitgroups=mg
 journalInferMarketPricesFromTransactions :: Journal -> Journal
 journalInferMarketPricesFromTransactions j =
   j{jinferredmarketprices =
-       dbg4 "jinferredmarketprices" .
+       dbg4With (("jinferredmarketprices:\n"<>) . showMarketPrices) $
        map priceDirectiveToMarketPrice .
        concatMap postingPriceDirectivesFromCost $
        journalPostings j
@@ -939,32 +979,23 @@ journalInferMarketPricesFromTransactions j =
 journalToCost :: ConversionOp -> Journal -> Journal
 journalToCost cost j@Journal{jtxns=ts} = j{jtxns=map (transactionToCost cost) ts}
 
+-- | Identify and tag (1) equity conversion postings and (2) postings which have (or could have ?) redundant costs.
+-- And if the addcosts flag is true, also add any costs which can be inferred from equity conversion postings.
+-- This is always called before transaction balancing to tag the redundant-cost postings so they can be ignored.
+-- With --infer-costs, it is called again after transaction balancing (when it has more information to work with) to infer costs from equity postings.
+-- See transactionTagCostsAndEquityAndMaybeInferCosts for more details, and hledger manual > Cost reporting for more background.
+journalTagCostsAndEquityAndMaybeInferCosts :: Bool -> Bool -> Journal -> Either String Journal
+journalTagCostsAndEquityAndMaybeInferCosts verbosetags addcosts j = do
+  let conversionaccts = journalConversionAccounts j
+  ts <- mapM (transactionTagCostsAndEquityAndMaybeInferCosts verbosetags addcosts conversionaccts) $ jtxns j
+  return j{jtxns=ts}
+
 -- | Add equity postings inferred from costs, where needed and possible.
 -- See hledger manual > Cost reporting.
 journalInferEquityFromCosts :: Bool -> Journal -> Journal
-journalInferEquityFromCosts verbosetags j = journalMapTransactions (transactionAddInferredEquityPostings verbosetags equityAcct) j
-  where
-    equityAcct = journalConversionAccount j
-
--- | Add costs inferred from equity conversion postings, where needed and possible.
--- See hledger manual > Cost reporting.
-journalInferCostsFromEquity :: Journal -> Either String Journal
-journalInferCostsFromEquity j = do
-  ts <- mapM (transactionInferCostsFromEquity False conversionaccts) $ jtxns j
-  return j{jtxns=ts}
-  where conversionaccts = journalConversionAccounts j
-
--- XXX duplication of the above
--- | Do just the internal tagging that is normally done by journalInferCostsFromEquity,
--- identifying equity conversion postings and, in particular, postings which have redundant costs.
--- Tagging the latter is useful as it allows them to be ignored during transaction balancedness checking.
--- And that allows journalInferCostsFromEquity to be postponed till after transaction balancing,
--- when it will have more information (amounts) to work with.
-journalMarkRedundantCosts :: Journal -> Either String Journal
-journalMarkRedundantCosts j = do
-  ts <- mapM (transactionInferCostsFromEquity True conversionaccts) $ jtxns j
-  return j{jtxns=ts}
-  where conversionaccts = journalConversionAccounts j
+journalInferEquityFromCosts verbosetags j =
+  journalMapTransactions (transactionInferEquityPostings verbosetags equityAcct) j
+  where equityAcct = journalBaseConversionAccount j
 
 -- -- | Get this journal's unique, display-preference-canonicalised commodities, by symbol.
 -- journalCanonicalCommodities :: Journal -> M.Map String CommoditySymbol
@@ -980,28 +1011,28 @@ journalMarkRedundantCosts j = do
 
 -- -- | Get this amount's commodity and any commodities referenced in its price.
 -- amountCommodities :: Amount -> [CommoditySymbol]
--- amountCommodities Amount{acommodity=c,aprice=p} =
+-- amountCommodities Amount{acommodity=c,acost=p} =
 --     case p of Nothing -> [c]
---               Just (UnitPrice ma)  -> c:(concatMap amountCommodities $ amounts ma)
---               Just (TotalPrice ma) -> c:(concatMap amountCommodities $ amounts ma)
+--               Just (UnitCost ma)  -> c:(concatMap amountCommodities $ amounts ma)
+--               Just (TotalCost ma) -> c:(concatMap amountCommodities $ amounts ma)
 
--- | Get an ordered list of amounts in this journal which can
--- influence canonical amount display styles. Those amounts are, in
--- the following order:
+-- | Get an ordered list of amounts in this journal which can influence
+-- canonical amount display styles (excluding the ones in commodity directives). 
+-- They are, in the following order:
 --
 -- * amounts in market price (P) directives (in parse order)
--- * posting amounts in transactions (in parse order)
+-- * posting amounts and optionally cost amounts (in parse order)
 -- * the amount in the final default commodity (D) directive
 --
--- Transaction price amounts (posting amounts' aprice field) are not included.
---
-journalStyleInfluencingAmounts :: Journal -> [Amount]
-journalStyleInfluencingAmounts j =
+journalStyleInfluencingAmounts :: Bool -> Journal -> [Amount]
+journalStyleInfluencingAmounts includecost j =
   dbg7 "journalStyleInfluencingAmounts" $
   catMaybes $ concat [
    [mdefaultcommodityamt]
   ,map (Just . pdamount) $ jpricedirectives j
-  ,map Just . concatMap (amountsRaw . pamount) $ journalPostings j
+  ,map Just $ if includecost
+    then journalPostingAndCostAmounts j
+    else concatMap amountsRaw $ journalPostingAmounts j
   ]
   where
     -- D's amount style isn't actually stored as an amount, make it into one
@@ -1033,7 +1064,7 @@ journalStyleInfluencingAmounts j =
 -- * posting amounts in transactions (in parse order)
 --
 -- Transaction price amounts, which may be embedded in posting amounts
--- (the aprice field), are left intact but not traversed/processed.
+-- (the acost field), are left intact but not traversed/processed.
 --
 -- traverseJournalAmounts :: Applicative f => (Amount -> f Amount) -> Journal -> f Journal
 -- traverseJournalAmounts f j =
@@ -1118,17 +1149,27 @@ pivotAccount :: Text -> Posting -> Text
 pivotAccount fieldortagname p =
   T.intercalate ":" [pivotComponent x p | x <- T.splitOn ":" fieldortagname]
 
+-- | Get the value of the given field or tag for this posting.
+-- "comm" and "cur" are accepted as synonyms meaning the commodity symbol.
+-- Pivoting on an unknown field or tag, or on commodity when there are multiple commodities, returns "".
+-- Pivoting on a tag when there are multiple values for that tag, returns the first value.
 pivotComponent :: Text -> Posting -> Text
 pivotComponent fieldortagname p
-  |                           fieldortagname == "acct"        = paccount p
-  | Just t <- ptransaction p, fieldortagname == "code"        = tcode t
-  | Just t <- ptransaction p, fieldortagname == "desc"        = tdescription t
-  | Just t <- ptransaction p, fieldortagname == "description" = tdescription t  -- backward compatible with 1.30 and older
-  | Just t <- ptransaction p, fieldortagname == "payee"       = transactionPayee t
-  | Just t <- ptransaction p, fieldortagname == "note"        = transactionNote t
-  | Just t <- ptransaction p, fieldortagname == "status"      = T.pack . show . tstatus $ t
-  | Just (_, value) <- postingFindTag fieldortagname p        = value
-  | otherwise                                                 = ""
+  | fieldortagname == "code",        Just t <- ptransaction p = tcode t
+  | fieldortagname `elem` descnames, Just t <- ptransaction p = tdescription t
+  | fieldortagname == "payee",       Just t <- ptransaction p = transactionPayee t
+  | fieldortagname == "note",        Just t <- ptransaction p = transactionNote t
+  | fieldortagname == "status",      Just t <- ptransaction p = T.pack . show . tstatus $ t
+  | fieldortagname == "acct"        = paccount p
+  | fieldortagname `elem` commnames = case map acommodity $ amounts $ pamount p of [s] -> s; _ -> unknown
+  | fieldortagname == "amt"         = case amounts $ pamount p of [a] -> T.pack $ show $ aquantity a; _ -> unknown
+  | fieldortagname == "cost"        = case amounts $ pamount p of [a@Amount{acost=Just _}] -> T.pack $ lstrip $ showAmountCost a; _ -> unknown
+  | Just (_, tagvalue) <- postingFindTag fieldortagname p = tagvalue
+  | otherwise = unknown
+  where
+    descnames = ["desc", "description"]   -- allow "description" for hledger <=1.30 compat
+    commnames = ["cur","comm"]            -- allow either; cur is the query prefix, comm is more consistent
+    unknown   = ""
 
 postingFindTag :: TagName -> Posting -> Maybe (TagName, TagValue)
 postingFindTag tagname p = find ((tagname==) . fst) $ postingAllTags p
@@ -1199,7 +1240,7 @@ samplejournalMaybeExplicit explicit = nulljournal
          {jtxns = [
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 01 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1216,7 +1257,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1233,7 +1274,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 02,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1250,7 +1291,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 03,
              tdate2=Nothing,
              tstatus=Cleared,
@@ -1267,7 +1308,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 10 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1283,7 +1324,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 12 31,
              tdate2=Nothing,
              tstatus=Unmarked,
