@@ -35,9 +35,9 @@ import Control.Applicative ((<|>))
 import Data.Function ((&), on)
 import Data.List (partition, intercalate, sortBy)
 import Data.List.Extra (nubSortBy)
-import qualified Data.Map as M
-import qualified Data.Set as S
-import qualified Data.Text as T
+import Data.Map qualified as M
+import Data.Set qualified as S
+import Data.Text qualified as T
 import Data.Time.Calendar (Day, fromGregorian)
 import Data.MemoUgly (memo)
 import GHC.Generics (Generic)
@@ -48,7 +48,8 @@ import Hledger.Data.Types
 import Hledger.Data.Amount
 import Hledger.Data.Dates (nulldate)
 import Text.Printf (printf)
-import Data.Decimal (decimalPlaces, roundTo)
+import Data.Decimal (decimalPlaces, roundTo, Decimal)
+import Data.Word (Word8)
 
 
 ------------------------------------------------------------------------------
@@ -94,7 +95,9 @@ journalPriceOracle :: Bool -> Journal -> PriceOracle
 journalPriceOracle infer Journal{jpricedirectives, jinferredmarketprices} =
   let
     declaredprices = map priceDirectiveToMarketPrice jpricedirectives
-    inferredprices = if infer then jinferredmarketprices else []
+    inferredprices =
+      (if infer then jinferredmarketprices else [])
+      & dbg2Msg ("use prices inferred from costs? " <> if infer then "yes" else "no")
     makepricegraph = memo $ makePriceGraph declaredprices inferredprices
   in
     memo $ uncurry3 $ priceLookup makepricegraph
@@ -112,13 +115,13 @@ priceDirectiveToMarketPrice PriceDirective{..} =
 -- The price's display precision will be set to show all significant
 -- decimal digits; or if they seem to be infinite, defaultPrecisionLimit.
 amountPriceDirectiveFromCost :: Day -> Amount -> Maybe PriceDirective
-amountPriceDirectiveFromCost d amt@Amount{acommodity=fromcomm, aquantity=n} = case aprice amt of
-    Just (UnitPrice u)           -> Just $ pd{pdamount=u}
-    Just (TotalPrice t) | n /= 0 -> Just $ pd{pdamount=u}
-      where u = amountSetFullPrecisionOr Nothing $ divideAmount n t
+amountPriceDirectiveFromCost d amt@Amount{acommodity=fromcomm, aquantity=n} = case acost amt of
+    Just (UnitCost u)           -> Just $ pd{pdamount=u}
+    Just (TotalCost t) | n /= 0 -> Just $ pd{pdamount=u}
+      where u = amountSetFullPrecisionUpTo Nothing $ divideAmount n t
     _                            -> Nothing
   where
-    pd = PriceDirective{pddate = d, pdcommodity = fromcomm, pdamount = nullamt}
+    pd = PriceDirective{pdsourcepos=nullsourcepos, pddate=d, pdcommodity=fromcomm, pdamount=nullamt}
 
 ------------------------------------------------------------------------------
 -- Converting things to value
@@ -209,7 +212,7 @@ amountValueAtDate priceoracle styles mto d a =
       -- set the display precision to match the internal precision (showing all digits),
       -- unnormalised (don't strip trailing zeros);
       -- but if it looks like an infinite decimal, limit the precision to 8.
-      & amountSetFullPrecisionOr Nothing
+      & amountSetFullPrecisionUpTo Nothing
       & dbg9With (lbl "calculated value".showAmount)
 
 -- | Calculate the gain of each component amount, that is the difference
@@ -261,7 +264,7 @@ priceLookup makepricegraph d from mto =
               ,pgEdgesRev=allprices
               ,pgDefaultValuationCommodities=defaultdests
               } =
-      traceOrLogAt 1 ("valuation date: "++show d) $ makepricegraph d
+      dbg1Msg ("valuation date: "++show d) $ makepricegraph d
     mto' = mto <|> mdefaultto
       where
         mdefaultto = dbg1 ("default valuation commodity for "++T.unpack from) $
@@ -273,14 +276,17 @@ priceLookup makepricegraph d from mto =
       Just to            ->
         -- We have a commodity to convert to. Find the most direct price available,
         -- according to the rules described in makePriceGraph.
-        let msg = printf "seeking %s to %s price" (showCommoditySymbol from) (showCommoditySymbol to)
-        in case 
-          (traceOrLogAt 2 (msg++" using forward prices") $ 
-            pricesShortestPath from to forwardprices)
-          <|> 
-          (traceOrLogAt 2 (msg++" using forward and reverse prices") $ 
-            pricesShortestPath from to allprices)
-        of
+        let
+          msg = printf "seeking %s to %s price" (showCommoditySymbol from) (showCommoditySymbol to)
+          prices =
+            (dbg2Msg (msg++" using forward prices") $
+             dbg2Msg ("forward prices:\n" <> showMarketPrices forwardprices) $
+             pricesShortestPath from to forwardprices)
+            <|>
+            (dbg2Msg (msg++" using forward and reverse prices") $
+             dbg2Msg ("forward and reverse prices:\n" <> showMarketPrices allprices) $
+             pricesShortestPath from to $ dbg5 "all forward and reverse prices" allprices)
+        in case prices of
           Nothing -> Nothing
           Just [] -> Nothing
           Just ps -> Just (mpto $ last ps, rate)
@@ -290,12 +296,17 @@ priceLookup makepricegraph d from mto =
                 -- aggregate all the prices into one
                 product rates
                 -- product (Decimal's Num instance) normalises, stripping trailing zeros.
-                -- Here we undo that (by restoring the old max precision with roundTo), 
-                -- so that amountValueAtDate can see the original internal precision,
-                -- to use as the display precision of calculated value amounts.
-                -- (This can add more than the original number of trailing zeros to some prices,
-                -- making them seem more precise than they were, but it seems harmless here.)
-                & roundTo (maximum $ map decimalPlaces rates)
+                -- But we want to preserve even those, since the number of decimal digits
+                -- here will guide amountValueAtDate in setting the Amount display precision later.
+                -- So we restore them. Or rather, we ensure as many decimal digits as the maximum seen among rates.
+                -- (Some prices might end up more precise than they were, but that seems harmless here.)
+                & setMinDecimalPlaces (maximum $ map decimalPlaces rates)
+
+-- Ensure this Decimal has at least this many decimal places, adding trailing zeros if necessary.
+setMinDecimalPlaces :: Word8 -> Decimal -> Decimal
+setMinDecimalPlaces n d
+  | decimalPlaces d < n = roundTo n d  -- too few, add some zeros
+  | otherwise           = d            -- more than enough, keep as-is
 
 tests_priceLookup =
   let
@@ -326,7 +337,7 @@ data PriceGraph = PriceGraph {
     -- ^ The date on which these prices are in effect.
   ,pgEdges :: [Edge]
     -- ^ "Forward" exchange rates between commodity pairs, either
-    --   declared by P directives or inferred from transaction prices,
+    --   declared by P directives or (with --infer-market-prices) inferred from costs,
     --   forming the edges of a directed graph.  
   ,pgEdgesRev :: [Edge]
     -- ^ The same edges, plus any additional edges that can be
@@ -334,6 +345,7 @@ data PriceGraph = PriceGraph {
     --
     --   In both of these there will be at most one edge between each
     --   directed pair of commodities, eg there can be one USD->EUR and one EUR->USD.
+    --
   ,pgDefaultValuationCommodities :: M.Map CommoditySymbol CommoditySymbol
     -- ^ The default valuation commodity for each source commodity.
     --   These are used when a valuation commodity is not specified
@@ -352,9 +364,8 @@ pricesShortestPath :: CommoditySymbol -> CommoditySymbol -> [Edge] -> Maybe Path
 pricesShortestPath start end edges =
   -- at --debug=2 +, print the pretty path and also the detailed prices
   let label = printf "shortest path from %s to %s: " (showCommoditySymbol start) (showCommoditySymbol end) in
-  fmap (dbg2With (("price chain:\n"++).pshow)) $ 
-  dbg2With ((label++).(maybe "none found" (pshowpath ""))) $
-
+  fmap (dbg2With (("price chain:\n"++).showMarketPrices)) $
+  dbg2With ((label++).(maybe "none" (pshowpath ""))) $
   find [([],edges)]
 
   where
@@ -364,13 +375,8 @@ pricesShortestPath start end edges =
       case concatMap extend paths of
         [] -> Nothing 
         _ | pathlength > maxpathlength -> 
-          -- XXX This is unusual:
-          -- 1. A warning, not an error, which we usually avoid
-          -- 2. Not a debug message (when triggered, we always print it)
-          -- 3. Printed either to stdout or (eg in hledger-ui) to the debug log file.
-          -- This is the only place we use traceOrLog like this.
-          -- Also before 1.32.2, traceOrLog was doing the opposite of what it should [#2134].
-          traceOrLog ("gave up searching for a price chain at length "++show maxpathlength++", please report a bug")
+          -- Print a non-fatal warning to stderr, something we usually avoid.
+          warn ("gave up searching for a price chain at length "++show maxpathlength++", please report a bug")
           Nothing
           where 
             pathlength = 2 + maybe 0 (length . fst) (headMay paths)
