@@ -87,11 +87,17 @@ This is used by the import command.
 module Hledger.Read (
 
   -- * Journal files
-  PrefixedFilePath,
   defaultJournal,
+  defaultJournalSafely,
+  defaultJournalWith,
+  defaultJournalWithSafely,
   defaultJournalPath,
+  defaultJournalPathSafely,
   requireJournalFileExists,
   ensureJournalFileExists,
+  journalEnvVar,
+  -- journalEnvVar2,
+  journalDefaultFilename,
 
   -- * Journal parsing
   runExceptT,
@@ -102,14 +108,15 @@ module Hledger.Read (
 
   -- * Easy journal parsing
   readJournal',
+  readJournal'',
   readJournalFile',
   readJournalFiles',
   orDieTrying,
 
   -- * Misc
-  journalStrictChecks,
   saveLatestDates,
   saveLatestDatesForFiles,
+  isWindowsUnsafeDotPath,
 
   -- * Re-exported
   JournalReader.tmpostingrulep,
@@ -125,8 +132,8 @@ module Hledger.Read (
 ) where
 
 --- ** imports
-import qualified Control.Exception as C
-import Control.Monad (unless, when)
+import Control.Exception qualified as C
+import Control.Monad (unless, when, forM, (<=<))
 import "mtl" Control.Monad.Except (ExceptT(..), runExceptT, liftEither)
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import Data.Default (def)
@@ -137,18 +144,17 @@ import Data.Maybe (catMaybes, fromMaybe)
 import Data.Ord (comparing)
 import Data.Semigroup (sconcat)
 import Data.Text (Text)
-import qualified Data.Text as T
-import qualified Data.Text.IO as T
+import Data.Text qualified as T
+import Data.Text.IO qualified as T
 import Data.Time (Day)
-import Safe (headDef)
-import System.Directory (doesFileExist, getHomeDirectory)
+import Safe (headDef, headMay)
+import System.Directory (doesFileExist)
 import System.Environment (getEnv)
-import System.Exit (exitFailure)
 import System.FilePath ((<.>), (</>), splitDirectories, splitFileName, takeFileName)
 import System.Info (os)
-import System.IO (hPutStr, stderr)
+import System.IO (Handle, hPutStrLn, stderr)
 
-import Hledger.Data.Dates (getCurrentDay, parsedateM, showDate)
+import Hledger.Data.Dates (getCurrentDay, parsedate, showDate)
 import Hledger.Data.Types
 import Hledger.Read.Common
 import Hledger.Read.InputOptions
@@ -159,7 +165,8 @@ import Hledger.Read.RulesReader (tests_RulesReader)
 -- import Hledger.Read.TimeclockReader (tests_TimeclockReader)
 import Hledger.Utils
 import Prelude hiding (getContents, writeFile)
-import Hledger.Data.JournalChecks (journalCheckAccounts, journalCheckCommodities)
+import Hledger.Data.JournalChecks (journalStrictChecks)
+import Text.Printf (printf)
 
 --- ** doctest setup
 -- $setup
@@ -168,40 +175,75 @@ import Hledger.Data.JournalChecks (journalCheckAccounts, journalCheckCommodities
 --- ** journal reading
 
 journalEnvVar           = "LEDGER_FILE"
-journalEnvVar2          = "LEDGER"
 journalDefaultFilename  = ".hledger.journal"
 
--- | Read the default journal file specified by the environment, or raise an error.
+-- | Read the default journal file specified by the environment, 
+-- with default input options, or raise an error.
 defaultJournal :: IO Journal
-defaultJournal = defaultJournalPath >>= runExceptT . readJournalFile definputopts >>= either error' return  -- PARTIAL:
+defaultJournal = defaultJournalSafely >>= either error' return -- PARTIAL:
 
--- | Get the default journal file path specified by the environment.
--- Like ledger, we look first for the LEDGER_FILE environment
--- variable, and if that does not exist, for the legacy LEDGER
--- environment variable. If neither is set, or the value is blank,
--- return the hard-coded default, which is @.hledger.journal@ in the
--- users's home directory (or in the current directory, if we cannot
--- determine a home directory).
+-- | Like defaultJournal, but return an error message instead of raising an error.
+defaultJournalSafely :: IO (Either String Journal)
+defaultJournalSafely = defaultJournalWithSafely definputopts
+
+-- | Like defaultJournal, but use the given input options.
+defaultJournalWith :: InputOpts -> IO Journal
+defaultJournalWith iopts = defaultJournalWithSafely iopts >>= either error' return -- PARTIAL:
+
+-- | Like defaultJournalWith, but return an error message instead of raising an error.
+defaultJournalWithSafely :: InputOpts -> IO (Either String Journal)
+defaultJournalWithSafely iopts = (do
+  f <- defaultJournalPath
+  runExceptT $ readJournalFile iopts f
+  )
+  `C.catches` [  -- XXX
+     C.Handler (\(e :: C.ErrorCall)   -> return $ Left $ show e)
+    ,C.Handler (\(e :: C.IOException) -> return $ Left $ show e)
+    ]
+
+-- | Get the default journal file path, and check that it exists; or raise an error.
+--
+-- This looks for the LEDGER_FILE environment variable, like Ledger.
+-- The value should be a file path; ~ at the start is supported, meaning user's home directory.
+-- The value can also be a glob pattern, for convenience; if so we consider only the first matched file.
+-- If no such file exists, an error is raised.
+--
+-- If LEDGER_FILE is unset or set to the empty string, we return a default file path:
+-- @.hledger.journal@ in the user's home directory.
+-- Or if we can't find the user's home directory, in the current directory.
+-- If this default file doesn't exist, an error is raised.
+--
 defaultJournalPath :: IO String
 defaultJournalPath = do
-  s <- envJournalPath
-  if null s then defpath else return s
-    where
-      envJournalPath =
-        getEnv journalEnvVar
-         `C.catch` (\(_::C.IOException) -> getEnv journalEnvVar2
-                                            `C.catch` (\(_::C.IOException) -> return ""))
-      defpath = do
-        home <- getHomeDirectory `C.catch` (\(_::C.IOException) -> return "")
-        return $ home </> journalDefaultFilename
+  ledgerfile <- getEnv journalEnvVar `C.catch` (\(_::C.IOException) -> return "")
+  if null ledgerfile
+  then do
+    homedir <- fromMaybe "" <$> getHomeSafe
+    let defaultfile = homedir </> journalDefaultFilename
+    exists <- doesFileExist defaultfile
+    if exists then return defaultfile
+    -- else error' $ "LEDGER_FILE is unset and \"" <> defaultfile <> "\" was not found"
+    else error' $ "neither LEDGER_FILE nor \"" <> defaultfile <> "\" were found"
+  else do
+    mf <- headMay <$> expandGlob "." ledgerfile `C.catch` (\(_::C.IOException) -> return [])
+    case mf of
+      Just f -> return f
+      Nothing -> error' $ "LEDGER_FILE points to nonexistent \"" <> ledgerfile <> "\""
 
--- | A file path optionally prefixed by a reader name and colon
--- (journal:, csv:, timedot:, etc.).
-type PrefixedFilePath = FilePath
+-- | Like defaultJournalPath, but return an error message instead of raising an error.
+defaultJournalPathSafely :: IO (Either String String)
+defaultJournalPathSafely = (do
+  f <- defaultJournalPath
+  return $ Right f
+  )
+  `C.catches` [
+     C.Handler (\(e :: C.ErrorCall)   -> return $ Left $ show e)
+    ,C.Handler (\(e :: C.IOException) -> return $ Left $ show e)
+    ]
 
 -- | @readJournal iopts mfile txt@
 --
--- Read a Journal from some text, with strict checks if enabled,
+-- Read a Journal from some handle, with strict checks if enabled,
 -- or return an error message.
 --
 -- The reader (data format) is chosen based on, in this order:
@@ -213,18 +255,25 @@ type PrefixedFilePath = FilePath
 -- - a file extension in @mfile@
 --
 -- If none of these is available, or if the reader name is unrecognised,
--- we use the journal reader (for predictability).
+-- the journal reader is used.
 --
-readJournal :: InputOpts -> Maybe FilePath -> Text -> ExceptT String IO Journal
-readJournal iopts@InputOpts{strict_} mpath txt = do
+-- If a file path is not provided, "-" is assumed (and may appear in error messages,
+-- `files` output etc, where it will be a slight lie: it will mean "not from a file",
+-- not necessarily "from standard input".
+--
+readJournal :: InputOpts -> Maybe FilePath -> Handle -> ExceptT String IO Journal
+readJournal iopts@InputOpts{strict_, _defer} mpath hdl = do
   let r :: Reader IO = fromMaybe JournalReader.reader $ findReader (mformat_ iopts) mpath
   dbg6IO "readJournal: trying reader" (rFormat r)
-  j <- rReadFn r iopts (fromMaybe "(string)" mpath) txt
-  when strict_ $ liftEither $ journalStrictChecks j
+  j <- rReadFn r iopts (fromMaybe "-" mpath) hdl
+  when (strict_ && not _defer) $ liftEither $ journalStrictChecks j
   return j
 
 -- | Read a Journal from this file, or from stdin if the file path is -,
 -- with strict checks if enabled, or return an error message.
+-- XXX or, calls error if the file does not exist.
+--
+-- (Note strict checks are disabled temporarily here when this is called by readJournalFiles).
 -- The file path can have a READER: prefix.
 --
 -- The reader (data format) to use is determined from (in priority order):
@@ -237,15 +286,14 @@ readJournal iopts@InputOpts{strict_} mpath txt = do
 -- generation, a rules file for converting CSV data, etc.
 --
 -- If using --new, and if latest-file writing is enabled in input options,
--- and after passing strict checks if enabled, a .latest.FILE file will be created/updated
--- (for the main file only, not for included files),
--- to remember the latest transaction date (and how many transactions on this date)
--- successfully read.
+-- and not deferred by readJournalFiles, and after passing strict checks if enabled,
+-- a .latest.FILE file will be created/updated (for the main file only, not for included files),
+-- to remember the latest transaction date processed.
 --
 readJournalFile :: InputOpts -> PrefixedFilePath -> ExceptT String IO Journal
-readJournalFile iopts@InputOpts{new_, new_save_} prefixedfile = do
+readJournalFile iopts@InputOpts{new_, new_save_, _defer} prefixedfile = do
   (j, mlatestdates) <- readJournalFileAndLatestDates iopts prefixedfile
-  when (new_ && new_save_) $ liftIO $
+  when (new_ && new_save_ && not _defer) $ liftIO $
     case mlatestdates of
       Nothing                        -> return ()
       Just (LatestDatesForFile f ds) -> saveLatestDates ds f
@@ -260,11 +308,11 @@ readJournalFileAndLatestDates iopts prefixedfile = do
     (mfmt, f) = splitReaderPrefix prefixedfile
     iopts' = iopts{mformat_=asum [mfmt, mformat_ iopts]}
   liftIO $ requireJournalFileExists f
-  t <-
-    traceOrLogAt 6 ("readJournalFile: "++takeFileName f) $
-    liftIO $ readFileOrStdinPortably f
+  h <-
+    dbg6Msg ("readJournalFile: "++takeFileName f) $
+    liftIO $ openFileOrStdin f
     -- <- T.readFile f  -- or without line ending translation, for testing
-  j <- readJournal iopts' (Just f) t
+  j <- readJournal iopts' (Just f) h
   if new_ iopts
     then do
       ds <- liftIO $ previousLatestDates f
@@ -275,9 +323,6 @@ readJournalFileAndLatestDates iopts prefixedfile = do
 
 -- | Read a Journal from each specified file path (using @readJournalFile@) 
 -- and combine them into one; or return the first error message.
--- Strict checks, if enabled, are deferred till the end.
--- Writing .latest files, if enabled, is also deferred till the end,
--- and happens only if strict checks pass.
 --
 -- Combining Journals means concatenating them, basically.
 -- The parse state resets at the start of each file, which means that
@@ -285,36 +330,39 @@ readJournalFileAndLatestDates iopts prefixedfile = do
 -- They do affect included child files though.
 -- Also the final parse state saved in the Journal does span all files.
 --
+-- Strict checks, if enabled, are temporarily deferred until all files are read,
+-- to ensure they see the whole journal, and/or to avoid redundant work.
+-- (Some checks, like assertions and ordereddates, might still be doing redundant work ?)
+--
+-- Writing .latest files, if enabled, is also deferred till the end,
+-- and is done only if strict checks pass.
+--
 readJournalFiles :: InputOpts -> [PrefixedFilePath] -> ExceptT String IO Journal
-readJournalFiles iopts@InputOpts{strict_,new_,new_save_} prefixedfiles = do
-  let iopts' = iopts{strict_=False, new_save_=False}
+readJournalFiles iopts@InputOpts{strict_, new_, new_save_} prefixedfiles = do
+  let iopts' = iopts{_defer=True}
   (j, latestdatesforfiles) <-
-    traceOrLogAt 6 ("readJournalFiles: "++show prefixedfiles) $
+    dbg6Msg ("readJournalFiles: "++show prefixedfiles) $
     readJournalFilesAndLatestDates iopts' prefixedfiles
   when strict_ $ liftEither $ journalStrictChecks j
   when (new_ && new_save_) $ liftIO $ saveLatestDatesForFiles latestdatesforfiles
   return j
 
--- The implementation of readJournalFiles, but with --new, 
--- also returns the latest transaction date(s) read in each file.
--- Used by the import command, to save those at the end.
+-- The implementation of readJournalFiles.
+-- With --new, it also returns the latest transaction date(s) read in each file
+-- (used by the import command).
 readJournalFilesAndLatestDates :: InputOpts -> [PrefixedFilePath] -> ExceptT String IO (Journal, [LatestDatesForFile])
 readJournalFilesAndLatestDates iopts pfs = do
   (js, lastdates) <- unzip <$> mapM (readJournalFileAndLatestDates iopts) pfs
   return (maybe def sconcat $ nonEmpty js, catMaybes lastdates)
 
--- | Run the extra -s/--strict checks on a journal,
--- returning the first error message if any of them fail.
-journalStrictChecks :: Journal -> Either String ()
-journalStrictChecks j = do
-  journalCheckAccounts j
-  journalCheckCommodities j
-
--- | An easy version of 'readJournal' which assumes default options, and fails
--- in the IO monad.
-readJournal' :: Text -> IO Journal
+-- | An easy version of 'readJournal' which assumes default options, and fails in the IO monad.
+readJournal' :: Handle -> IO Journal
 readJournal' = orDieTrying . readJournal definputopts Nothing
 
+-- | An even easier version of readJournal' which takes a 'Text' instead of a 'Handle'.
+readJournal'' :: Text -> IO Journal
+readJournal'' = readJournal' <=< inputToHandle
+
 -- | An easy version of 'readJournalFile' which assumes default options, and fails
 -- in the IO monad.
 readJournalFile' :: PrefixedFilePath -> IO Journal
@@ -331,29 +379,28 @@ readJournalFiles' = orDieTrying . readJournalFiles definputopts
 orDieTrying :: MonadIO m => ExceptT String m a -> m a
 orDieTrying a = either (liftIO . fail) return =<< runExceptT a
 
--- | If the specified journal file does not exist (and is not "-"),
--- give a helpful error and quit.
+-- | If the specified journal file does not exist (and is not "-"), call error with an informative message.
+-- (Using "journal file" generically here; it could be in any of hledger's supported formats.)
 requireJournalFileExists :: FilePath -> IO ()
 requireJournalFileExists "-" = return ()
 requireJournalFileExists f = do
   exists <- doesFileExist f
-  unless exists $ do  -- XXX might not be a journal file
-    hPutStr stderr $ "The hledger journal file \"" <> f <> "\" was not found.\n"
-    hPutStr stderr "Please create it first, eg with \"hledger add\" or a text editor.\n"
-    hPutStr stderr "Or, specify an existing journal file with -f or LEDGER_FILE.\n"
-    exitFailure
+  unless exists $ error' $ unlines
+    [ "data file \"" <> f <> "\" was not found."
+    ,"Please create it first, eg with \"hledger add\" or a text editor."
+    ,"Or, specify an existing data file with -f or $LEDGER_FILE."
+    ]
 
 -- | Ensure there is a journal file at the given path, creating an empty one if needed.
 -- On Windows, also ensure that the path contains no trailing dots
 -- which could cause data loss (see 'isWindowsUnsafeDotPath').
 ensureJournalFileExists :: FilePath -> IO ()
 ensureJournalFileExists f = do
-  when (os/="mingw32" && isWindowsUnsafeDotPath f) $ do
-    hPutStr stderr $ "Part of file path \"" <> show f <> "\"\n ends with a dot, which is unsafe on Windows; please use a different path.\n"
-    exitFailure
+  when (os=="mingw32" && isWindowsUnsafeDotPath f) $
+    error' $ "Part of file path \"" <> show f <> "\"\n ends with a dot, which is unsafe on Windows; please use a different path.\n"
   exists <- doesFileExist f
   unless exists $ do
-    hPutStr stderr $ "Creating hledger journal file " <> show f <> ".\n"
+    hPutStrLn stderr $ "Creating hledger journal file " <> show f
     -- note Hledger.Utils.UTF8.* do no line ending conversion on windows,
     -- we currently require unix line endings on all platforms.
     newJournalContent >>= T.writeFile f
@@ -376,6 +423,7 @@ type LatestDates = [Day]
 
 -- The path of an input file, and its current "LatestDates".
 data LatestDatesForFile = LatestDatesForFile FilePath LatestDates
+  deriving Show
 
 -- | Get all instances of the latest date in an unsorted list of dates.
 -- Ie, if the latest date appears once, return it in a one-element list,
@@ -386,8 +434,10 @@ latestDates = {-# HLINT ignore "Avoid reverse" #-}
 
 -- | Save the given latest date(s) seen in the given data FILE,
 -- in a hidden file named .latest.FILE, creating it if needed.
+-- Unless no latest dates are provided, in which case do nothing.
 saveLatestDates :: LatestDates -> FilePath -> IO ()
-saveLatestDates dates f = T.writeFile (latestDatesFileFor f) $ T.unlines $ map showDate dates
+saveLatestDates dates f = when (not $ null dates) $
+  T.writeFile (latestDatesFileFor f) $ T.unlines $ map showDate dates
 
 -- | Save each file's latest dates.
 saveLatestDatesForFiles :: [LatestDatesForFile] -> IO ()
@@ -401,12 +451,15 @@ saveLatestDatesForFiles = mapM_ (\(LatestDatesForFile f ds) -> saveLatestDates d
 previousLatestDates :: FilePath -> IO LatestDates
 previousLatestDates f = do
   let latestfile = latestDatesFileFor f
-      parsedate s = maybe (fail $ "could not parse date \"" ++ s ++ "\"") return $
-                      parsedateM s
   exists <- doesFileExist latestfile
-  if exists
-  then traverse (parsedate . T.unpack . T.strip) . T.lines =<< readFileStrictly latestfile
-  else return []
+  t <- if exists then readFileStrictly latestfile else return T.empty
+  let nls = zip [1::Int ..] $ T.lines t
+  fmap catMaybes $ forM nls $ \(n,l) -> do
+    let s = T.unpack $ T.strip l
+    case (s, parsedate s) of
+      ("", _)       -> return Nothing
+      (_,  Nothing) -> error' (printf "%s:%d: invalid date: \"%s\"" latestfile n s)
+      (_,  Just d)  -> return $ Just d
 
 -- | Where to save latest transaction dates for the given file path.
 -- (.latest.FILE)
