@@ -44,13 +44,13 @@ import Control.Monad.Except (ExceptT, liftEither)
 import Control.Monad.State.Strict
 import Data.Char (isSpace)
 import Data.Text (Text)
-import qualified Data.Text as T
+import Data.Text qualified as T
 import Data.Time (Day)
 import Text.Megaparsec hiding (parse)
 import Text.Megaparsec.Char
 
 import Hledger.Data
-import Hledger.Read.Common hiding (emptyorcommentlinep)
+import Hledger.Read.Common
 import Hledger.Utils
 import Data.Decimal (roundTo)
 import Data.Functor ((<&>))
@@ -66,26 +66,24 @@ import Data.List (group)
 
 reader :: MonadIO m => Reader m
 reader = Reader
-  {rFormat     = "timedot"
+  {rFormat     = Timedot
   ,rExtensions = ["timedot"]
-  ,rReadFn     = parse
+  ,rReadFn     = handleReadFnToTextReadFn parse
   ,rParser    = timedotp
   }
 
 -- | Parse and post-process a "Journal" from the timedot format, or give an error.
 parse :: InputOpts -> FilePath -> Text -> ExceptT String IO Journal
-parse iopts fp t = initialiseAndParseJournal timedotp iopts fp t
+parse iopts fp t = initialiseAndParseJournal (timedotp iopts) iopts fp t
                    >>= liftEither . journalApplyAliases (aliasesFromOpts iopts)
                    >>= journalFinalise iopts fp t
 
 --- ** utilities
 
-traceparse, traceparse' :: String -> TextParser m ()
-traceparse  = const $ return ()
-traceparse' = const $ return ()
--- for debugging:
--- traceparse  s = traceParse (s++"?")
--- traceparse' s = trace s $ return ()
+-- Trace parser state above a certain --debug level ?
+tracelevel = 9
+dp :: String -> JournalParser m ()
+dp = if tracelevel >= 0 then lift . dbgparse tracelevel else const $ return ()
 
 --- ** parsers
 {-
@@ -108,14 +106,13 @@ Org headings before the first date line are ignored, regardless of content.
 
 timedotfilep = timedotp -- XXX rename export above
 
-timedotp :: JournalParser m ParsedJournal
-timedotp = preamblep >> many dayp >> eof >> get
+timedotp :: InputOpts -> JournalParser m ParsedJournal
+timedotp _ = preamblep >> many dayp >> eof >> get
 
 preamblep :: JournalParser m ()
 preamblep = do
-  lift $ traceparse "preamblep"
-  many $ notFollowedBy datelinep >> (lift $ emptyorcommentlinep "#;*")
-  lift $ traceparse' "preamblep"
+  dp "preamblep"
+  void $ many $ notFollowedBy datelinep >> (lift $ emptyorcommentlinep2 "#;*")
 
 -- | Parse timedot day entries to multi-posting time transactions for that day.
 -- @
@@ -126,11 +123,13 @@ preamblep = do
 -- @
 dayp :: JournalParser m ()
 dayp = label "timedot day entry" $ do
-  lift $ traceparse "dayp"
+  dp "dayp"
   pos <- getSourcePos
   (date,desc,comment,tags) <- datelinep
+  dp "dayp1"
   commentlinesp
-  ps <- (many $ timedotentryp <* commentlinesp) <&> concat
+  dp "dayp2"
+  ps <- (many $ dp "dayp3" >> timedotentryp <* commentlinesp) <&> concat
   endpos <- getSourcePos
   let t = txnTieKnot $ nulltransaction{
     tsourcepos   = (pos, endpos),
@@ -145,7 +144,7 @@ dayp = label "timedot day entry" $ do
 
 datelinep :: JournalParser m (Day,Text,Text,[Tag])
 datelinep = do
-  lift $ traceparse "datelinep"
+  dp "datelinep"
   lift $ optional orgheadingprefixp
   date <- datep
   desc <- T.strip <$> lift descriptionp
@@ -156,16 +155,15 @@ datelinep = do
 -- or org headlines which do not start a new day.
 commentlinesp :: JournalParser m ()
 commentlinesp = do
-  lift $ traceparse "commentlinesp"
-  void $ many $ try $ lift $ emptyorcommentlinep "#;"
+  dp "commentlinesp"
+  void $ many $ try $ lift $ emptyorcommentlinep2 "#;"
 
 -- orgnondatelinep :: JournalParser m ()
 -- orgnondatelinep = do
---   lift $ traceparse "orgnondatelinep"
+--   dp "orgnondatelinep"
 --   lift orgheadingprefixp
 --   notFollowedBy datelinep
 --   void $ lift restofline
---   lift $ traceparse' "orgnondatelinep"
 
 orgheadingprefixp = skipSome (char '*') >> skipNonNewlineSpaces1
 
@@ -175,10 +173,10 @@ orgheadingprefixp = skipSome (char '*') >> skipNonNewlineSpaces1
 -- @
 timedotentryp :: JournalParser m [Posting]
 timedotentryp = do
-  lift $ traceparse "timedotentryp"
+  dp "timedotentryp"
   notFollowedBy datelinep
   lift $ optional $ choice [orgheadingprefixp, skipNonNewlineSpaces1]
-  a <- modifiedaccountnamep
+  a <- modifiedaccountnamep False
   lift skipNonNewlineSpaces
   taggedhours <- lift durationsp
   (comment0, tags0) <-
@@ -225,7 +223,7 @@ durationsp =
 -- @
 numericquantityp :: TextParser m Hours
 numericquantityp = do
-  -- lift $ traceparse "numericquantityp"
+  -- dp "numericquantityp"
   (q, _, _, _) <- numberp Nothing
   msymbol <- optional $ choice $ map (string . fst) timeUnits
   skipNonNewlineSpaces
@@ -257,7 +255,7 @@ timeUnits =
 -- @
 dotquantityp :: TextParser m Hours
 dotquantityp = do
-  -- lift $ traceparse "dotquantityp"
+  -- dp "dotquantityp"
   char '.'
   dots <- many (oneOf ['.', ' ']) <&> filter (not.isSpace)
   return $ fromIntegral (1 + length dots) / 4
@@ -267,7 +265,7 @@ dotquantityp = do
 -- ignoring any interspersed spaces after the first letter.
 letterquantitiesp :: TextParser m [(Hours, TagValue)]
 letterquantitiesp =
-  -- dbg "letterquantitiesp" $ 
+  -- dp "letterquantitiesp"
   do
     letter1 <- letterChar
     letters <- many (letterChar <|> spacenonewline) <&> filter (not.isSpace)
@@ -276,19 +274,3 @@ letterquantitiesp =
           | t@(c:_) <- group $ sort $ letter1:letters
           ]
     return groups
-
--- | XXX new comment line parser, move to Hledger.Read.Common.emptyorcommentlinep
--- Parse empty lines, all-blank lines, and lines beginning with any of the provided
--- comment-beginning characters.
-emptyorcommentlinep :: [Char] -> TextParser m ()
-emptyorcommentlinep cs =
-  label ("empty line or comment line beginning with "++cs) $ do
-    traceparse "emptyorcommentlinep" -- XXX possible to combine label and traceparse ?
-    skipNonNewlineSpaces
-    void newline <|> void commentp
-    traceparse' "emptyorcommentlinep"
-    where
-      commentp = do
-        choice (map (some.char) cs)
-        takeWhileP Nothing (/='\n') <* newline
-
