@@ -28,9 +28,10 @@ module Hledger.Reports.ReportTypes
 , flatDisplayName
 , treeDisplayName
 
+, prrShowDebug
 , prrFullName
 , prrDisplayName
-, prrDepth
+, prrIndent
 , prrAdd
 ) where
 
@@ -44,6 +45,8 @@ import GHC.Generics (Generic)
 import Hledger.Data
 import Hledger.Query (Query)
 import Hledger.Reports.ReportOptions (ReportOpts)
+import Data.Text qualified as T
+import Data.List (intercalate)
 
 type Percentage = Decimal
 
@@ -119,6 +122,14 @@ instance HasAmounts b => HasAmounts (PeriodicReportRow a b) where
      ,prrAverage=styleAmounts styles $ prrAverage r
      }
 
+prrShowDebug :: PeriodicReportRow DisplayName MixedAmount -> String
+prrShowDebug (PeriodicReportRow dname amts _tot _avg) =
+  unwords [
+    T.unpack $ displayFull dname,
+    "",
+    intercalate " | " $ map showMixedAmount amts
+    ]
+
 -- | Add two 'PeriodicReportRows', preserving the name of the first.
 prrAdd :: Semigroup b => PeriodicReportRow a b -> PeriodicReportRow a b -> PeriodicReportRow a b
 prrAdd (PeriodicReportRow n1 amts1 t1 a1) (PeriodicReportRow _ amts2 t2 a2) =
@@ -132,8 +143,10 @@ zipWithPadded _ []     bs     = bs
 
 -- | Figure out the overall date span of a PeriodicReport
 periodicReportSpan :: PeriodicReport a b -> DateSpan
-periodicReportSpan (PeriodicReport [] _ _)       = DateSpan Nothing Nothing
-periodicReportSpan (PeriodicReport colspans _ _) = DateSpan (fmap Exact . spanStart $ head colspans) (fmap Exact . spanEnd $ last colspans)
+periodicReportSpan (PeriodicReport colspans _ _) =
+  case colspans of
+    []  -> DateSpan Nothing Nothing
+    s:_ -> DateSpan (Exact <$> spanStart s) (Exact <$> spanEnd (last colspans))
 
 -- | Map a function over the row names.
 prMapName :: (a -> b) -> PeriodicReport a c -> PeriodicReport b c
@@ -198,36 +211,39 @@ data CBCSubreportSpec a = CBCSubreportSpec
   }
 
 
--- | A full name, display name, and depth for an account.
+-- | The number of indentation steps with which to display a report item.
+-- 0 means no indentation. 1 means one indent step, which is normally rendered
+-- as two spaces in text output, or two no-break spaces in csv/html output.
+type NumberOfIndents = Int
+
+-- | A full name, display name, and indent level for an account.
 data DisplayName = DisplayName
-    { displayFull :: AccountName
-    , displayName :: AccountName
-    , displayDepth :: Int
+    { displayFull   :: AccountName
+    , displayName   :: AccountName
+    , displayIndent :: NumberOfIndents
     } deriving (Show, Eq, Ord)
 
 instance ToJSON DisplayName where
     toJSON = toJSON . displayFull
     toEncoding = toEncoding . displayFull
 
--- | Construct a flat display name, where the full name is also displayed at
--- depth 1
+-- | Construct a display name for a list report, where full names are shown unindented.
 flatDisplayName :: AccountName -> DisplayName
-flatDisplayName a = DisplayName a a 1
+flatDisplayName a = DisplayName a a 0
 
--- | Construct a tree display name, where only the leaf is displayed at its
--- given depth
+-- | Construct a display name for a tree report, where leaf names (possibly prefixed by
+-- boring parents) are shown indented).
 treeDisplayName :: AccountName -> DisplayName
 treeDisplayName a = DisplayName a (accountLeafName a) (accountNameLevel a)
 
--- | Get the full, canonical, name of a PeriodicReportRow tagged by a
--- DisplayName.
+-- | Get the full canonical account name from a PeriodicReportRow containing a DisplayName.
 prrFullName :: PeriodicReportRow DisplayName a -> AccountName
 prrFullName = displayFull . prrName
 
--- | Get the display name of a PeriodicReportRow tagged by a DisplayName.
+-- | Get the account display name from a PeriodicReportRow containing a DisplayName.
 prrDisplayName :: PeriodicReportRow DisplayName a -> AccountName
 prrDisplayName = displayName . prrName
 
--- | Get the display depth of a PeriodicReportRow tagged by a DisplayName.
-prrDepth :: PeriodicReportRow DisplayName a -> Int
-prrDepth = displayDepth . prrName
+-- | Get the indent level from a PeriodicReportRow containing a DisplayName.
+prrIndent :: PeriodicReportRow DisplayName a -> Int
+prrIndent = displayIndent . prrName
